import { isPromise } from "@implab/core-amd/safe";
import { Unsubscribable, Producer, Observable, Subscribable } from "./observable/interfaces";
import { ObservableImpl, fuse } from "./observable/ObservableImpl";

export const isUnsubscribable = (v: unknown): v is Unsubscribable =>
    v !== null && v !== undefined && typeof (v as Unsubscribable).unsubscribe === "function";

export const isSubscribable = <T = unknown>(v: unknown): v is Subscribable<T> =>
    v !== null && v !== undefined && typeof (v as Subscribable<unknown>).subscribe === "function";

export const observe = <T>(producer: Producer<T>): Observable<T> => new ObservableImpl<T>(fuse(producer));

/** Converts an array to the observable sequence of its elements. */
export const ofArray = <T>(items: T[]): Observable<T> => new ObservableImpl<T>(
    ({ next, complete }) => (
        items.forEach(next),
        complete()
    )
);

/** Converts a subscribable to the observable */
export const ofSubscribable = <T>(subscribable: Subscribable<T>) =>
    observe<T>(sink => {
        const subscription = subscribable.subscribe(sink);
        return () => subscription.unsubscribe();
    });

const of1 = <T>(item: T | PromiseLike<T>) => observe<T>(
    ({ next, error, complete }) =>
        isPromise(item) ?
            void item.then(
                v => (next(v), complete()),
                error
            ) :
            (next(item), complete())
);

/** Converts a list of parameter values to the observable sequence. The
 * order of elements in the list will be preserved in the resulting sequence.
 */
export const of = <T>(...items: (T | PromiseLike<T>)[]) => items.length === 1 ?
    of1(items[0]) :
    observe<T>(
        ({ next, error, complete, isClosed }) => {
            const n = items.length;

            const _next = (start: number) => {
                if (start > 0 && isClosed()) // when resumed
                    return;

                for (let i = start; i < n; i++) {
                    const r = items[i];
                    if (isPromise(r)) {
                        r.then(v => (next(v), _next(i + 1)), error);
                        return; // suspend
                    } else {
                        next(r);
                    }
                }
                complete();
            };

            _next(0);
        }
    );

export const empty: Observable<never> = new ObservableImpl<never>(({ complete }) => complete());

export type * from "./observable/interfaces";